//! Wrapper around internal pipes fds.
//!
//! These pipe fds are valid across forks, and can be used for IPC between parent and non-exec()d child process forks.
use super::*;
use errno::Errno;
use std::{error, fmt};
use std::ops::Drop;
use std::io;

/// Kind of pipe error
#[derive(Debug)]
#[non_exhaustive]
pub enum ErrorKind
{
    /// Failed to create pipes
    Create,
    /// Unknown error
    Unknown,
}

/// A raw unix pipe
// TODO: Change uses of `i32` for socket fds. Or change this to a struct (see below.)
pub type RawPipe = i32;

/// Create a raw pipe pair.
pub(crate) fn unix_pipe() -> Result<(i32, i32), Error>
{
    use libc::pipe;
    let mut pipe_fds: [libc::c_int; 2] = [-1;2];
    if unsafe{pipe(pipe_fds.as_mut_ptr())} == 0 {
	Ok((i32::from(pipe_fds[0]), i32::from(pipe_fds[1])))
    } else {
	Err(ErrorKind::Create.into_error())
    }
}

//TODO: Factor out the `i32` raw sockets in these types below into a seperate `RawPipe` struct, which has the Drop impl and has functions for transfering ownership of the raw socket (it should hide the raw fd when not explicitly needed ideally.)
//      This will prevent us from having to ensure we `forget()` the outer type each time we want to move the inner one around without closing it.

/// A writer for a bi-directinoal unix pipe
///
/// Created by splitting `Pipe`, data written to this is available to be read from its `ReadHalf` counterpart.
#[derive(Debug, PartialEq, Eq)]
pub struct WriteHalf(i32);

/// A reader for a bi-directional unix pipe.
///
/// Created by splitting `Pipe`, data read from this is data that was sent to its `WriteHalf` counterpart.
#[derive(Debug, PartialEq, Eq)]
pub struct ReadHalf(i32);

/// A bi-drectional unix pipe
///
/// Data written to the pipe is available to be read from it (unless the pipe is mismatched.)
#[derive(Debug, PartialEq, Eq)]
pub struct Pipe
{
    tx: i32,
    rx: i32,
}

// Make sure when operating on consumers of Pipe of Read/WriteHalf, the pipe structures are `mem::forget`ed before returning, since the Drop impl of those types closes the raw socket.
impl Pipe
{
    /// Split this pipe into a read and write half.
    ///
    /// Data written to the `WriteHalf` is sent to the `ReadHalf` (unless the pipe is mismatched.)
    #[inline] pub fn split(self) -> (WriteHalf, ReadHalf)
    {
	let rv = (WriteHalf(self.tx), ReadHalf(self.rx));	
	std::mem::forget(self);
	rv
    }

    /// Create a `Pipe` from two halves.
    ///
    /// # Mismatched halves
    /// The halves do not need to have originated from the same `Pipe`.
    /// If they are not, then writing to the `Pipe` will send the data to its original receiver, and reading from it will take the data from its original sender.
    #[inline] pub fn unsplit(tx: WriteHalf, rx: ReadHalf) -> Self
    {
	let v = Self::new_from_raw((tx.0, rx.0));
	std::mem::forget((tx, rx));
	v
    }

    /// Create a new `Pipe` from a pair of raw file descriptors.
    ///
    /// # Safety
    /// Does not check that the integers provided are valid and open file descriptors.
    #[inline(always)] fn new_from_raw((tx, rx): (i32, i32)) -> Self
    {
	Self{tx,rx}
    }
    
    /// Create a new `Pipe` from a pair of raw file descriptors.
    ///
    /// # Safety
    /// The caller must check that `tx` and `rx` are valid, open file descriptors
    // TODO: When writing the async version make sure to state that the fds need to be non-blocking.
    #[inline] pub unsafe fn from_raw(tx: i32, rx: i32) -> Self
    {
	Self::new_from_raw((tx, rx))
    }

    /// Transfer the ownership of the socket's inner Tx and Rx fds.
    pub fn into_raw(self) -> (i32, i32)
    {
	let txrx = (self.tx, self.rx);
	std::mem::forget(self);
	txrx
    }

    /// Create a new `Pipe` from two new file descriptors that stream to eachother.
    ///
    /// Data written to this pipe will also be read from it.
    /// The pipe can be `split()`, and then `unsplit()` with a different split pipe's half to create a stream between those two pipes.
    #[inline] pub fn new() -> Result<Self, Error>
    {
	let sk = unix_pipe()?;
	Ok(Self::new_from_raw(sk))
    }
}

/// Create a pair of bi-directional linked pipes.
///
/// Writing into one of the pipes will send the data to the other one, and vise-versa.
pub fn pair() -> Result<(Pipe, Pipe), Error>
{
    let (a, b) = (Pipe::new()?, Pipe::new()?);
    let (atx, arx) = a.split();
    let (btx, brx) = b.split();
    Ok((
	Pipe::unsplit(atx, brx),
	Pipe::unsplit(btx, arx),
    ))
}

/// Write to a raw fd.
fn write_raw(sock: i32, data: &[u8]) -> Result<usize, Errno>
{
    match unsafe {libc::write(sock, data.as_ptr() as *const _, data.len())} {
	x if x < 0 => Err(Errno::last().unwrap_err()),
	x => Ok(usize::try_from(x).unwrap()),
    }
}

/// Read from a raw fd.
fn read_raw(sock: i32, data: &mut [u8]) -> Result<usize, Errno>
{   
    match unsafe {libc::read(sock, data.as_ptr() as *mut _, data.len())} {
	x if x < 0 => Err(Errno::last().unwrap_err()),
	x => Ok(usize::try_from(x).unwrap()),
    }
}

// There's nothing to flush, the fd isn't buffered.
// fn flush_raw(sock: i32) -> Result<(), Errno>
// {
//     libc::fsync(sock)
// }

impl io::Read for Pipe
{
    #[inline] fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
	Ok(read_raw(self.rx, buf)?)
    }
}
impl io::Write for Pipe
{
    #[inline] fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
	Ok(write_raw(self.tx, buf)?)
    }
    #[inline] fn flush(&mut self) -> io::Result<()> {
	Ok(())
    }
}

impl io::Read for ReadHalf
{
    #[inline] fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
	Ok(read_raw(self.0, buf)?)
    }
}
impl io::Write for WriteHalf
{
    #[inline] fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
	Ok(write_raw(self.0, buf)?)
    }
    #[inline] fn flush(&mut self) -> io::Result<()> {
	Ok(())
    }
}


impl Drop for ReadHalf
{
    fn drop(&mut self) {
	unsafe {
	    libc::close(self.0);
	}
    }
}
impl Drop for WriteHalf
{
    fn drop(&mut self) {
	unsafe {
	    libc::close(self.0);
	}
    }
}
impl Drop for Pipe
{
    fn drop(&mut self) {
	unsafe {
	    libc::close(self.tx);
	    libc::close(self.rx);
	}
    }
}

impl fmt::Display for ErrorKind
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	match self {
	    Self::Create => write!(f, "failed to create raw pipes"),
	    _ => write!(f, "unknown error"),
	}
    }
}

/// An error on a pipe operation
#[derive(Debug)]
pub struct Error(ErrorKind, Errno);

impl ErrorKind
{
    /// Consume into `Error` using the last `errno` value, if it's not set then return `val`.
    #[inline] pub(crate) fn try_into_error_or<T>(self, val: T) -> Result<T, Error>
    {
	Error::or_last(self, val)
    }    
    /// Consume into `Error` using the last `errno` value, if it's not set then return the result of `fun`.
    #[inline] pub(crate) fn try_into_error_or_with<F, T>(self, fun: F) -> Result<T, Error>
    where F: FnOnce() -> T
    {
	Error::or_last_with(self, fun)
    }
    /// Consume into `Error` using the last `errno` value.
    pub(crate) fn try_into_error(self) -> Result<(), Error>
    {
	Error::last(self)
    }

    /// Consume into an `Error` using the last `errno` value.
    ///
    /// # Panics
    /// If `errno` was success.
    #[inline] pub(crate) fn into_error(self) -> Error
    {
	Self::try_into_error(self).expect_err("Expected an error to be set in errno")
    }
}

impl Error
{
    /// Create a new error of this kind from the last set errno if it's not success.
    /// If it is, return `val`.
    #[inline] pub(crate) fn or_last<T>(err: ErrorKind, val: T) -> Result<T, Self>
    {
	Self::or_last_with(err, move || val)
    }
    /// Create a new error of this kind from the last set errno if it's not success.
    /// If it is, return the result of `fun`.
    pub(crate) fn or_last_with<F, T>(err: ErrorKind, fun: F) -> Result<T, Self>
    where F: FnOnce() -> T
    {
	Err(Self(err, match Errno::or_last_with(fun) {
	    Ok(v) => return Ok(v),
	    Err(e) => e,
	}))
    }
    /// Create a new error of this kind from the last set errno (if it's not success.)
    pub(crate) fn last(err: ErrorKind) -> Result<(), Self>
    {
	Err(Self(err, match Errno::last() {
	    Ok(()) => return Ok(()),
	    Err(e) => e,
	}))
    }

    /// Create a new error of this kind with this errno value
    #[inline] pub fn new(kind: ErrorKind, err: Errno) -> Self
    {
	Self(kind, err)
    }

    /// The kind of pipe error
    #[inline] pub fn kind(&self) -> &ErrorKind
    {
	&self.0
    }

    /// The errno value
    #[inline] pub fn error(&self) -> &Errno
    {
	&self.1
    }
}

impl error::Error for Error
{
    fn source(&self) -> Option<&(dyn error::Error + 'static)> {
	Some(&self.1)
    }
}
impl fmt::Display for Error
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	write!(f, "{}: {}", self.0, self.1)
    }
}

